#include "Cguide.h"
#include "Cguide_X.h"
#include "Cguide_globals.h"
#include "Cguide_std.proto"

/* test */

void extract_string(FILE *inf, char str[])
{
	char c;
	int i;

	while (isspace(c = getc(inf)))
		;
	i = 0;
	do {
		if (c != '\r')
			str[i++] = c;
	} while ((c = getc(inf)) != '\n');
	str[i++] = '\0';
}

void extract_id(FILE *inf, char id[])
{
	/* The id number of the article is extracted, and then padded with
	zeros so that it is of the form:

		[editor code] + [single char reality code] + [editor-article-number]
	
	where the editor code & the editor-article-number are specified by the
	constant CODE_LENGTH in the header, which is currently three.

	This was done so that, thinking ahead, I have allowed for 999 editors
	to edit 999 articles of each reality level each. If PGG ever has more
	articles than that I will happily revise this code length! */

	char temp[ID_LENGTH];
	char c;
	int i, j;

	/* initialise ID number to 0's */
	for (i = 0; i < ID_LENGTH; i++)
		id[i] = '0';

	/* skip leading whitespace */
	while (isspace(c = getc(inf)))
		;
	/* get the editor code */
	i = 0;
	temp[i++] = c;
	while (isdigit(temp[i++] = getc(inf)))
		;
	/* store the reality code (first non-digit) */
	id[CODE_LENGTH] = temp[--i];
	/* copy editor code into correct places */
	for (j = CODE_LENGTH - 1; i > 0; )
		id[j--] = temp[--i];
	/* get the editor-article-number */
	i = 0;
	while (isdigit(temp[i++] = getc(inf)))
		;
	/* store last char read */
	c = temp[--i];
	/* copy into correct places */
	for (j = 2*CODE_LENGTH; i > 0; )
		id[j--] = temp[--i];
	id[ID_LENGTH - 1] = '\0';
	/* move file pointer to start of next line */
	while (c != '\n')
		c = getc(inf);
}

void add_xref(char *title, XREF **top)
{
	XREF *current;

	if (*top == NULL) {
		*top = (XREF *)malloc(sizeof(XREF));
		(*top)->next = NULL;
		strcpy((*top)->xtitle, title);
	}
	else {
		current = *top;
		while (current->next != NULL)
			current = current->next;
		current->next = (XREF *)malloc(sizeof(XREF));
		current = current->next;
		current->next = NULL;
		strcpy(current->xtitle, title);
	}
}

void print_xrefs(XREF *top)
{
	XREF *current;
	if (top == NULL)
		return;
	current = top;
	do {
		printf("Cross-Reference: %s\n", current->xtitle);
		current = current->next;
	} while (current != NULL);
}

void add_index(FILE *inf, INDEX **top)
{
	INDEX *current;

	if (*top == NULL) {
		*top = (INDEX *)malloc(sizeof(INDEX));
		(*top)->next = NULL;
		extract_string(inf, (*top)->index_entry);
	}
	else {
		current = *top;
		while (current->next != NULL)
			current = current->next;
		current->next = (INDEX *)malloc(sizeof(INDEX));
		current = current->next;
		current->next = NULL;
		extract_string(inf, current->index_entry);
	}
}

void print_indexes(INDEX *top)
{
	INDEX *current;
	if (top == NULL)
		return;
	current = top;
	do {
		printf("Index: %s\n", current->index_entry);
		current = current->next;
	} while (current != NULL);
}

int count_entries(char *ifname)
{
	FILE *ifile;
	int c, last_c;
	int num_entries;

	if ((ifile = fopen(ifname, "r")) == NULL) {
		printf("Could not open file: %s\n");
		return(0);
	}
	/* count entries */
	num_entries = 0;
	last_c = '\n'; /* At start of file, pretend just found end of line */
	while(!feof(ifile)) {
		c = getc(ifile);
		if (c != EOF) {
			if (last_c != '\n') { /* only interested in a %t at start of line */
				last_c = c;
				continue;
			}
			else {
				if (c != '%') { /* if 1st char of line not '%', forget it */
					last_c = c;
					continue;
				}
				else {
				c = getc(ifile);
				if (c == TITLE_CHAR)
					num_entries++;
				last_c = c;
				}
			}
		}
	}
	fclose(ifile);
	return(num_entries);
}

GUIDE_ENTRY *create_new_entry()
{
	GUIDE_ENTRY *entry;
	entry = (GUIDE_ENTRY *)malloc(sizeof(GUIDE_ENTRY));
	entry->first_xref = NULL;
	entry->first_index = NULL;
	entry->article = NULL;
	return(entry);
}

int get_entries(GUIDE_ENTRY **entries, char *ifname, int ent_no)
{
	int c;
	char temp_str[LINE_LENGTH];
	int in_entry = FALSE;
	FILE *ifile;

	in_entry = FALSE;
	if ((ifile = fopen(ifname, "r")) == NULL) {
		printf("Error: unable to open file: %s\n", ifname);
		return(ent_no);
	}
	while(!feof(ifile)) {
		c = getc(ifile); /* get first char of next line */
		if (c != EOF) {
			if (c != '%')
				continue;
			else {
				c = getc(ifile);
				switch(c) {
				case TITLE_CHAR:
					if (entries[ent_no] == NULL)
						entries[ent_no] = create_new_entry();
					extract_string(ifile, entries[ent_no]->title);
					break;
                case SUMMARY_CHAR:
					extract_string(ifile, entries[ent_no]->summary);
					break;
                case AUTHOR_CHAR:
					extract_string(ifile, entries[ent_no]->author);
					break;
                case DATE_CHAR:
					extract_string(ifile, temp_str);
					entries[ent_no]->date = atol(temp_str);
					break;
				case NUMBER_CHAR:
					extract_id(ifile, entries[ent_no]->id);
					break;
				case XREF_CHAR:
					extract_string(ifile, temp_str);
					add_xref(temp_str, &entries[ent_no]->first_xref);
					break;
				case INDEX_CHAR:
					add_index(ifile, &entries[ent_no]->first_index);
					total_index_entries++;
					break;
				case DELIMIT_CHAR:
					if (in_entry)
						ent_no++;
					else
						entries[ent_no]->article_length = extract_text(ifile, &entries[ent_no]->article);
					in_entry = !in_entry;
					break;	
				default:
					break;
				}
			}
		}
	}
	fclose(ifile);
	return(ent_no);
}

int create_working_files(char *namebase, GUIDE_ENTRY **entries, int total_entries)
{
	char cguide_fname[FNAME_LENGTH];
	char entries_fname[FNAME_LENGTH];
	char index_fname[FNAME_LENGTH];
	char xref_fname[FNAME_LENGTH];
	FILE *cguide_file;
	FILE *entries_file;
	FILE *index_file;
	FILE *xref_file;
	long current_ef_pos; /* current position in entry file */
	XREF *current, *top;
	int i, j;

	strncpy(cguide_fname, namebase, 8);
	strcat(cguide_fname, ".CGD");
	strncpy(entries_fname, namebase, 8);
	strcat(entries_fname, ".ENT");
	strncpy(index_fname, namebase, 8);
	strcat(index_fname, ".IDX");
	strncpy(xref_fname, namebase, 8);
	strcat(xref_fname, ".XRF");

	if ((cguide_file = fopen(cguide_fname, "wb")) == NULL) {
		printf("Could not open .CGD file for writing.\n");
		return(FALSE);
	}
	if ((entries_file = fopen(entries_fname, "w")) == NULL) {
		printf("Could not open .ENT file for writing.\n");
		return(FALSE);
	}
	if ((index_file = fopen(index_fname, "wb")) == NULL) {
		printf("Could not open .IDX file for writing.\n");
		return(FALSE);
	}
	if ((xref_file = fopen(xref_fname, "wb")) == NULL) {
		printf("Could not open .XRF file for writing.\n");
		return(FALSE);
	}
	
	fwrite(&total_entries, sizeof(int), 1, cguide_file);
	for (i = 0; i < total_entries; i++) {
		current_ef_pos = ftell(entries_file);
		for (j = 0; j < entries[i]->article_length; j++)
			putc(entries[i]->article[j], entries_file);
		entries[i]->text_start = current_ef_pos;
		fwrite(entries[i], sizeof(GUIDE_ENTRY), 1, cguide_file);
		top = entries[i]->first_xref;
		if (top != NULL) {
			current = top;
			do {
				fwrite(&i, sizeof(int), 1, xref_file); 
				fwrite(current->xtitle, sizeof(char), LINE_LENGTH, xref_file); 
				current = current->next;
			} while (current != NULL);
		}
		fflush(cguide_file);
		fflush(entries_file);
		fflush(xref_file);
	}
	fwrite(&total_index_entries, sizeof(int), 1, index_file);
	fwrite(index_entries, sizeof(INDEX), total_index_entries, index_file);
	fclose(cguide_file);
	fclose(entries_file);
	fclose(index_file);
	fclose(xref_file);
	return(TRUE);
}

GUIDE_ENTRY **load_working_files(char *namebase, int *total_entries)
{
	GUIDE_ENTRY **entries;
	char cguide_fname[FNAME_LENGTH];
	char entries_fname[FNAME_LENGTH];
	char index_fname[FNAME_LENGTH];
	char xref_fname[FNAME_LENGTH];
	FILE *cguide_file;
	FILE *entries_file;
	FILE *index_file;
	FILE *xref_file;
	long current_ef_pos; /* current position in entry file */
	char temp_str[LINE_LENGTH];
	int i, j;

	strncpy(cguide_fname, namebase, 8);
	strcat(cguide_fname, ".CGD");
	strncpy(index_fname, namebase, 8);
	strcat(index_fname, ".IDX");
	strncpy(xref_fname, namebase, 8);
	strcat(xref_fname, ".XRF");

	if ((cguide_file = fopen(cguide_fname, "rb")) == NULL) {
		printf(".CGD file not found\n");
		return(NULL);
	}
	if ((index_file = fopen(index_fname, "rb")) == NULL) {
		printf(".IDX file not found\n");
		return(NULL);
	}
	if ((xref_file = fopen(xref_fname, "rb")) == NULL) {
		printf(".XRF file not found\n");
		return(NULL);
	}
	
	fread(total_entries, sizeof(int), 1, cguide_file);
	printf("Total Entries: %d\n", *total_entries);
	entries = (GUIDE_ENTRY **)malloc(*total_entries*sizeof(GUIDE_ENTRY *));
	for (i = 0; i < *total_entries; i++) {
		entries[i] = create_new_entry();
		fread(entries[i], sizeof(GUIDE_ENTRY), 1, cguide_file);
		/* fix up pointers */
		entries[i]->first_xref = NULL;
		entries[i]->first_index = NULL;
		entries[i]->article = NULL;
	}
	while(!feof(xref_file)) {
		fread(&i, sizeof(int), 1, xref_file);
		if((fread(temp_str, sizeof(char), LINE_LENGTH, xref_file)) != 0)
			if ((i >= 0) && (i < *total_entries))
				add_xref(temp_str, &entries[i]->first_xref);
	}
	fread(&total_index_entries, sizeof(int), 1, index_file);
	if (index_entries != NULL)
		free(index_entries);
	index_entries = (INDEX *)malloc(total_index_entries*sizeof(INDEX));
	fread(index_entries, sizeof(INDEX), total_index_entries, index_file);
	fclose(cguide_file);
	fclose(index_file);
	fclose(xref_file);
	if (toc != NULL)
		free(toc);
	toc = (int *)malloc((*total_entries)*sizeof(int));
	return(entries);
}

void update_working_files()
{
	int i, j;
	int num_ifiles;
	char **ifnames;
	int current_entry;
	INDEX *current_ndx, *top_ndx;

	total_entries = 0;
	total_index_entries = 0;
    ifnames = get_filenames(&num_ifiles);
    for (i = 0; i < num_ifiles; i++)
        total_entries += count_entries(ifnames[i]);
    printf("Total number of entries: %d\n\n", total_entries);
    entries = (GUIDE_ENTRY **)malloc(total_entries*sizeof(GUIDE_ENTRY *));
    for (i = 0; i < total_entries; i++)
        entries[i] = NULL;
    current_entry = 0;
    for (i = 0; i < num_ifiles; i++)
        current_entry = get_entries(entries, ifnames[i], current_entry);
	printf("All articles read in.\n");
    /* now all the articles are in memory */
	/* sort articles by title */
	qsort(entries, total_entries, sizeof(*entries), title_cmp);
	/* create the index */
	index_entries = (INDEX *)malloc(total_index_entries*sizeof(INDEX));
	j = 0;
	for (i = 0; ((i < total_entries) && (j < total_index_entries)); i++) {
		top_ndx = entries[i]->first_index;
		if (top_ndx != NULL) {
			current_ndx = top_ndx;
			do {
				strcpy(index_entries[j].title, entries[i]->title);
				strcpy(index_entries[j].index_entry, current_ndx->index_entry);
				j++;
				current_ndx = current_ndx->next;
			} while (current_ndx != NULL);
		}
	}
	qsort(index_entries,total_index_entries,sizeof(*index_entries),index_cmp);
    create_working_files("CGUIDE", entries, total_entries);
}

int title_cmp(const void *ent_1, const void *ent_2)
{
	return(strcasecmp((*((GUIDE_ENTRY **)(ent_1)))->title, (*((GUIDE_ENTRY **)(ent_2)))->title));
}

int index_cmp(const void *ent_1, const void *ent_2)
{
	return(strcasecmp((*((INDEX *)(ent_1))).index_entry, (*((INDEX *)(ent_2))).index_entry));
}

